#!/usr/bin/env python
version = "0.7.5"

# Copyright (c) 2007 Daniel Graham <daniel.graham@duke.edu>. All rights
# reserved.
# 
# License:
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version. [http://www.gnu.org/licenses/gpl.html]
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

import sys, datetime, re, os, glob, os.path, fnmatch, shutil
from optparse import OptionParser
fatal = False
try:
	import yaml
except:
	print """
Sorry, yeagtd requires (py)yaml --- see http://pyyaml.org/.
"""
	fatal = True
try:
	from dateutil.rrule import *
except:
	print """
Sorry, yeagtd requires dateutil --- see  http://labix.org/python-dateutil
"""
	fatal = True
if fatal:
	print "python", sys.version, 
	for item in sys.path:
		print item
	sys.exit()

config = """
# -------------------- Begin ~/.yeagtdrc ---------------------------
# The following, edited to suit, should be saved to ~/.yeagtdrc
# The directory containing the GTD project files
gtddir = "/Users/dag/Documents/GTD"
# Use this command to invoke an external editor. Will be called
# with a space-separated list of files to open for editing 
edit_cmd = "mate"
# Use this command for printing yeaGTD output saved in a temp file
print_cmd = "a2ps -B -R -f10 --columns=1"
# Number of backup copies to keep, n. [Backup first moves   
# .bk(n-1) to bk(n), then .bk(n-2) to .bk(n-1) and so forth
# until .bk1 is moved to .bk2. Then .yaml is copied to bk1.]
numbaks = 2
# Create reminders for Remind using the following template
# or, to avoid creating reminders, just omit this line 
remtemplate = 'REM %(d)s %(r)s MSG %%"%(p)s/%(c)s: %(a)s%%" %%1%%'
# 'MSG' could be replaced by 'RUN <your alert command>', e.g., 
# 'RUN wxremalert' if you have wxRemind installed.
# These replacements will occur as project actions are processed:
# %(d)s -> the date, e.g., 'Jun 17 2007'
# %(r)s -> the project *remind* setting , e.g., 'AT 14:30 +15 *5'
# %%    -> %
# %(p)s -> the project title (the *agenda* setting)
# %(c)s -> the action context (the *c* setting)
# %(a)s -> the action title (the *a* setting)
# Reminders will be written to 'reminders' in gtddir. Be sure to 
# 'INCLUDE' this file in your ~/.reminders file.
# --------------------- End ~/.yeagtdrc ----------------------------
"""


yeagtdrc = os.path.expanduser('~/.yeagtdrc')

print_file = os.path.expanduser('~/.yeagtdtmp')


if os.path.isfile(yeagtdrc):
	remtemplate = None
	execfile(yeagtdrc)
	# remtemplate will either still be None or will be something like
	# the following
	# remtemplate = 'REM %(d)s %(r)s RUN wxremalert %%"%(p)s/%(c)s %(a)s%%" %%1%%'
	# remhash will have keys and suitable values for the replacments
else:
	print """
Fatal Error: The required configuration file, %s, could not be found.
%s
""" % (yeagtdrc, config)
	sys.exit()

if remtemplate:
	remfile = os.path.join(gtddir, 'reminders')
	fo = open(remfile, 'w')
	fo.write("# DO NOT EDIT - THIS FILE WILL BE OVERWRITTEN!\n")
	fo.close()
oneday = datetime.timedelta(1)

def CheatSheet():
	print """Illustrative Project Files
==========================

    A non-repeating project                A repeating project
  -----------------------------         ----------------------------

    agenda: Fix light                   agenda: Repetitive Project

    tasks:                              due: 2007-05-18
    - a: Remove old halogen bulb
      c: Home                           repeat:
                                          byweekday: FR
    - a: Buy similar bulb                 frequency: WEEKLY
      c: Errands                          interval: 2

    - a: Install new bulb               tasks:
      c: Home                           - a: First task
                                          c: Home

                                        - a: Second task
                                          c: Car
Project File Details
====================

    agenda: agenda or project title. Required.

    begin:  beginning date at which tasks become available in date-time
            format, i.e., either YYYY-MM-DD or YYYY-MM-DD HH:MM:SS.

    days:   non-negative number of days in advance of due at which the
            project's tasks become available.

    due:    due date for completing all tasks in date-time format.

            At most two of {begin, due, days} should be provided. When
            two are provided then the third is deduced, i.e., begin =
            due - days; due = begin + days; days = due - begin.
            Otherwise days = 0. If only begin is given then due = begin.
            If only due is given then begin = due. Otherwise begin =
            today and due = today.

            If both begin and due are specified and begin is later than
            due, then both begin and due are reset internally to the
            minimum of begin and due and an error message is displayed.

    notes:  project notes. Default = None.

    remind: Time specification for users who have remind installed.
            Default = None. If specified, a reminder will be written
            to the file 'reminders' in the yeaGTD project directory if
            the relevant action is available today. E.g., remind = "AT
            14:30 +15 DURATION 0:45" would set up a reminder to trigger
            today 15 minutes before 2:30PM and again at 2:30PM. Remind
            would display this as an event today beginning at 2:30PM and
            ending 45 minutes later with the message "agenda/context:
            act". For this to function properly, you must include the
            file 'reminders' from your yeaGTD project directory in your
            ~/.reminders file and yeagtd.py must, of course, be run
            before the reminder will be available to remind. The "AT
            HH:MM" is required - the rest of the string is optional.

    series: True/False. Default = True.

            If true, treat project actions as series by default. If
            false, treat project actions as parallel by default. See the
            p setting under actions and the discussion of Available
            Actions below.

    tasks:  A list of the actions required to complete the project. Each
            action has the following possible fields.

         a: action title. Required.

         n: notes. Default = None.

         b: date action becomes available in date-time format.

         c: user specified context. Default = '???'. Other possibilities
         include 'Office', 'Home', 'Errands', et cetera. Actions with the
         default category, "???", are combined with actions belonging to the
         Inbox project and listed first in all reports as a reminder to set
         contexts for all your actions and to "empty your inbox".

         d: non-negative number of days in advance of due at which the action
         becomes available. At most one of {b, d } should be specified. When
         one is specified the other is deduced, i.e., d = due - b or b = due -
         d. If neither is specified, then b = begin and d = due - begin (=
         days).

         f: date action was (last) finished in date-time format. Default
         = None.

         p: alphabetic parallel character or None. Default = None. See the
         setting for series above and discussion of available actions below.

    repeat: Optional. When included the following parameters can be
            specified.

         frequency: DAILY, WEEKLY, MONTHLY OR YEARLY. Required.

         interval: the number of periods between repetitions, e.g., a project
         with frequency = 'WEEKLY' and interval = 2 would repeat every other
         week. Default = 1.

         byweekday: the weekday or sequence of weekdays on which the project
         should repeat. Default = (MO, TU, WE, TH, FR, SA, SU), i.e., any day
         of the week. Alternatively, weekdays can be specified using integers
         with MO = 1.

         bymonthday: an integer or sequence of integer month days on which the
         project should repeat. Default = (1, ..., 31), i.e., any day of the
         month.

         byweek: an integer, or a sequence of integer weeks on which the
         project should repeat where the first week of the year is that
         containing at least four days of the new year. Default = (1,...,53),
         i.e., any week of the year.

         bymonth: an integer or sequence of integer months on which the
         project should repeat. Default = (1, ..., 12), i.e., any month of the
         year.

         count: the number of repetitions. The default is to repeat
         indefinitely.

         until: repetitions end after this date in date-time format. The
         default is to repeat indefinitely.

         next: 'keep', 'skip' or 'restart'. Default = 'keep'. When recurrent
         projects are finished on time, the value of next has no effect. When
         the project is finished after the due date, the value of next has the
         following consequences.

            next = 'keep'. The next date from the recurrence rule becomes the
            end date for the next copy just as if the current copy had been
            completed on time.

            next = 'skip'. The next date from the recurrence rule that falls
            after the actual completion date becomes the next end date.
            Effectively, copies with due dates that fall after the current due
            date and the current finish date are skipped.

            next = 'restart'. The starting date for the recurrence rule is
            reset to the actual completion date and the next date from the
            recurrence rule becomes the next end date.

Examples of recurrence rules
============================

    Once every other week:              Monthly on the 2nd and 15th:
       frequency:  WEEKLY                  frequency: MONTHLY
       interval: 2                         bymonthday: (2,15)

    The first Monday in every month:    Monthly on the first and last days:
       frequency: MONTHLY                  frequency: MONTHLY
       interval: 1                         bymonthday: (1,-1)
       byweekday: MO
                                        30 days from the last completion:
    The third Monday in every month:       frequency: DAILY
       frequency: MONTHLY                  interval: 30
       interval: 1                         next: restart
       byweekday: MO(+3)
                                        Every August 13th which is a Friday:
    The last Monday in every month:        frequency: YEARLY
       frequency: MONTHLY                  bymonth: 8
       interval: 1                         bymonthday: 13
       byweekday: MO(-1)                   byweekday: FR

    Weekly on Tu and Th for 5 weeks:    The first Tuesday after a Monday in
       frequency: WEEKLY                November every 4 years:
       interval: 2                         frequency: YEARLY
       byweekday: (TU,TH)                  interval: 4
       count: 10                           bymonth: 11
                                           byweekday: TU
                                           bymonthday: (2,3,4,5,6,7,8)"""
	sys.exit()

def filelist():
	files = []
	for path, subdirs, names in os.walk(gtddir):
		for name in names:
			if fnmatch.fnmatch(name, '*.yaml'):
				files.append(os.path.join(path,name))
	return files

def edit():
	files = filelist()
	# files = glob.glob(os.path.join(gtddir,"*.yaml"))
	for i in range(1, len(files)):
		print "%3d %s" % (i, files[i-1])
	if len(files) > 0:
		print """
Enter a space separated list of the numbers files to open 
for editing or an '*' to open all files: """,
		res = raw_input()
		if res == '*':
			editlist = files
		else:
			editlist = []
			nums = map(int, res.split())
			if len(nums) > 0:
				changes = []
				for num in nums:
					editlist.append(files[num-1])
			else:
				print "Canceled"
		if len(editlist) > 0:
			os.system("%s %s" % (edit_cmd, ' '.join(editlist)))
	else:
		print "There are no available actions to mark finished."

def backup(option, opt, value, parser):
	files = filelist()
	for file in files:
		pathname, ext = os.path.splitext(file)
		path, name = os.path.split(pathname)
		bakdir = os.path.join(path,'backup')
		if not os.path.isdir(bakdir):
			os.mkdir(bakdir)
		bak = os.path.join(bakdir, "%s.bk1" % name)
		for i in range(1, numbaks):
			baknum = numbaks - i
			nextnum = baknum + 1
			
			bakname = os.path.join(bakdir,"%s.bk%d" % (name, baknum))
			if os.path.exists(bakname):
				nextname = os.path.join(bakdir,"%s.bk%d" % (name, nextnum))
				shutil.move(bakname, nextname)
		shutil.copy(file, bak)
		logaction(file, "backed up")
	sys.exit()

def logaction(file, logentry):
	"""docstring for logaction"""
	pathname, ext = os.path.splitext(file)
	path, name = os.path.split(pathname)
	logdir = os.path.join(path,'log')
	if not os.path.isdir(logdir):
		os.mkdir(logdir)
	logfile = os.path.join(logdir, "%s.log" % name)
	now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	fo = open(logfile, 'a')
	fo.write("%s: %s\n" % (now, logentry))
		
def inbox():
	args = sys.argv[2:]
	if len(args) > 0:
		addstr = ' '.join(args)
		fo = open(os.path.join(gtddir,"Inbox.yaml"), "a")
		fo.write("-  t: %s\n" % addstr)
		fo.close()
	else:
		print "usage: yeagtd -a String to add to inbox"

def topic():
	if len(sys.argv) >= 3:
		str = sys.argv[2]
		sansext = os.path.splitext(str)[0]
		agenda = os.path.split(sansext)[1]
		# Dump the initial path sep if it is given
		if os.path.isabs(sansext):
			sansext = sansext[1:]
		path = os.path.join(gtddir, "%s.yaml" % sansext)
		pathexists = os.path.exists(path)
		if pathexists:
			print "%s already exists. Existing file will be opened for editing." % path, 
		else:
			# create the directory if necessary and the file
			dirname = os.path.dirname(path)
			direxists = os.path.isdir(dirname)
			if not direxists:
				try:
					os.mkdir(dirname)
					direxists = True
				except:
					print "Could not create '%s'. [At most one sub-directory can be created for a new project.]" % dirname
			if direxists:
				fo = open(path, "w")
				fo.write("agenda: %s" % agenda)
				fo.close()
				pathexists = True
		if pathexists:
			os.system("%s %s" % (edit_cmd, path))	
	else:
		print "usage: yeagtd -t <relative path to new project file>"


def loadData():
	files = filelist()
	data = {}
	for file in files:
		project = loadFileData(file)
		data[file] = project
	return data
	
def loadFileData(file):
	fo = open(file, 'r')
	project = yaml.load(fo)
	fo.close()
	return project

def saveFileData(file, project):
	output = yaml.dump(project, default_flow_style = False)
	fo = open(file, 'w')
	fo.write(output)
	fo.close()

def valueordefault(dict, key, default=None):
	if dict.has_key(key):
		return dict[key]
	else:
		return default

def dt(dtobj):
	# Hack to remove leading zeros from datetime components
	parts = map(int, dtobj.strftime("%Y,%m,%d,%H,%M,%S").split(','))
	parts = map(str, parts)
	return ", ".join(parts)
	
def markdone(file,actionnum):
	"""docstring for markdone"""
	project = loadFileData(file)
	if project.has_key('tasks'):
		if len(project['tasks']) >= actionnum:
			act = project['tasks'][actionnum - 1]['a'] 
			project['tasks'][actionnum - 1]['f'] = \
			 	datetime.datetime.now()
			saveFileData(file, project)
			logaction(file, "set 'f' for '%s'" % act)

def recorddone(option, opt, value, parser):
	data = loadData()
	fileinfo = parseData(data, 'c', 'a', 'm')
	if len(fileinfo) > 0:
		print """
Enter a space separated list of actions to mark done
using the current date and time as the time stamp: """,
		res = raw_input()
		nums = map(int, res.split())
		if len(nums) > 0:
			changes = []
			for num in nums:
				file, index = fileinfo[num-1]
				markdone(file, index)
			data = loadData()
			fileinfo = parseData(data, 'c', 'a', 'm')
		else:
			print "Canceled"
	else:
		print "There are no available actions to mark finished."
	sys.exit()

def parseData(data, sort, filter, export, ret=False):
	printoutput = False
	mark = False
	fileinfo = []
	delimiter = ''
	if export == 'p':
		printoutput = True
	elif export == 'c':
		delimiter = ','
	elif export == 't':
		delimiter = '\t'		
	elif export == 'm':
		mark = True
		filter = 'a'		
	output = ''

	cwidth = 7
	actwidth = 34
	agendawidth = 10

	mintime = datetime.time.min
	
	# FIXME: columns for delimited output- not used
	delcols = ['project-file', 'agenda', 'action-number', 'context', 'action-begin', 'action-due', 'action-finished', 'finished-or-due', 'remaining-days', 'action']
		
	sortopts = {
		'p' : "project",
		'c'	: "context",
		'd'	: "date",
		}
	
	filteropts = { 
		'a' : "available", 
		'r' : "remaining", 
		'c' : "completed",
		'n'	: "all",
		}


	if sortopts.has_key(sort):
		view = sortopts[sort]
	else:
		print "'%s' is an unrecognized option for -s" % sort
		sys.exit()
	if filteropts.has_key(filter):
		filter = filteropts[filter]
	else:
		print "'%s' is an unrecognized option for -f" % filter
		sys.exit()
	
	datefmt = "%b %d"
	remfmt = "%Y,%b,%d"
	delfmt = "%Y-%m-%d"
	today = datetime.datetime.now()
	acts = []
	inbox = []
	errors = []
	projects = data.keys()
	for file in projects:
		proj = data[file]
		pdone = 0
		pavail = 0
		pother = 0
		unfinished = False
		repeating = False
		action_num = 0
		begindueset = True
		
		try:
			agenda = proj['agenda'] # Raise an exception if key is missing 
		except:
			print "Exception raised processing", file
		due = valueordefault(proj,'due')
		begin = valueordefault(proj, 'begin')
		days = valueordefault(proj, 'days')
		remind = valueordefault(proj, 'remind')
		lastdone = valueordefault(proj, 'lastdone')
		lastdue = valueordefault(proj, 'lastdue')
		if days:
			days = datetime.timedelta(max(int(days), 0))
		if begin and type(begin) == 'datetime.date':
			begin = datetime.datetime.combine(begin, mintime)
		if due and type(due) == 'datetime.date':
			due = datetime.datetime.combine(due, mintime)
		if begin and due and days and days != due - begin:
			errors.append("Error: begin, due and days specified for %s.\n\tOnly 2 of the 3 parameters should be given." % (agenda))
		elif begin and due:
			if begin > due:
				errors.append("Error: b = %s > d = %s for %s." %
				 	(begin.strftime(datefmt), 
					due.strftime(datefmt), agenda))
				begin = due = min(begin,due)
			else:
				days = due - begin
		elif begin and days:
			due = begin + days
		elif due and days:
			begin = due - days
		elif begin:
			due = begin
			days = datetime.timedelta(0)
		elif due:
			begin = due
			days = datetime.timedelta(0)
		else:
			begin = due = today
			begindueset = False
			days = datetime.timedelta(0)
		# make date into datetimes
		if type(begin) != type(today):
			begin = datetime.datetime.combine(begin, mintime)
		if type(due) != type(today):
			due = datetime.datetime.combine(due, mintime)
		
		# series = proj.setdefault('series', True)
		series = valueordefault(proj, 'series', True)
		if len(agenda) > agendawidth:
			agendat = "%*s..." % (agendawidth-3,
				agenda[:agendawidth-3])
		else:
			agendat = agenda

		if proj.has_key('repeat'):
			repeating = True
			# The parts for rrule with period the key for FREQ
			repeat = proj['repeat']
			next = valueordefault(repeat, 'next','keep')
			ignore = ['frequency', 'next']
			frequency = repeat['frequency']
			rule = [frequency]			
			for key in repeat.keys():
				if key not in ignore:
					rule.append("%s=%s" % (key, repeat[key]))
			rule = ', '.join(rule)
			rstr = "rrule(%s, dtstart=datetime.datetime(%s))" \
				% (rule, dt(due))
			r = eval(rstr)
			if not lastdue or not lastdone:
				lastdue = lastdone = due - oneday
				proj['lastdone'] = lastdone
				proj['lastdue'] = lastdue
				saveFileData(file, proj)
			if lastdue:
				if type(lastdue) != type(today):
					lastdue = datetime.datetime.combine(
						lastdue, mintime)
			if lastdone:
				if type(lastdone) != type(today):
					lastdone = datetime.datetime.combine(
						lastdone, mintime)
			lastcompleted = lastdone
			if next == 'restart':
				rstr = "rrule(%s, dtstart=datetime.datetime(%s))" \
					% (rule, dt(lastdone))
				r = eval(rstr)
				due = r.after(lastdone)
			elif next == 'skip':
				due = r.after(lastdone)
			elif next == 'keep':
				due = r.after(lastdue)
			begin = due - days
				
		if proj.has_key('tasks') and proj['tasks'] and \
		 	len(proj['tasks']) > 0:
			num_actions = len(proj['tasks'])
			plist = []
			remplist = []
			ai = 0
			for act in proj['tasks']:
				a = act['a'] 
				f = valueordefault(act,'f')
				if f and type(f) != type(today):
					f = datetime.datetime.combine(f, mintime)
				if f:
					if repeating:
						 if lastdone:
							if f <= lastdone:
								unfinished = True
							else:
								lastcompleted = max(f, lastcompleted)
				else:
					unfinished = True
				d = valueordefault(act,'d')
				b = valueordefault(act,'b')
				if b:
					b = datetime.datetime.combine(b, mintime)
				if d:
					d = datetime.timedelta(max(int(d), 0))

				if b and b > due and begindueset:
					errors.append(
					"Error: b = %s > due = %s for %s / %s"\
					 % (b.strftime(datefmt), due.strftime(datefmt),
					 	agenda, a))
					b = due

				# At this point, begin, due and today are all defined
				bset = True
				if begindueset:
					if not d:
						if b:
							d = due - b
						else:
							d = days
					if not b:
						if d:
							b = due - d
						else:
							b = begin
				else:
					if not d:
						d = datetime.timedelta(0)
					if not b:
						b = today
						bset = False

				p = valueordefault(act,'p', '')
				if act.has_key('c') and act['c'] == None:
					c = act['c'] = '???'
				else:
					c = act.setdefault('c', '???')
				if len(c) > cwidth:
					ct = "%-*s..." % (cwidth-3, c[:cwidth-3])
				else:
					ct = c
				n = valueordefault(act, 'n')

				if f == None:
					complete = False
				elif lastdone:
					if f <= lastdone:
						complete = False
					else:
						complete = True
				else:
					complete = True 

				# Use ddate for sorting and display
				if f and complete:
					ddate = f
				elif begindueset:
					ddate = due
				elif bset:
					ddate = b
				else:
					ddate = today
				
				# Use remdays to display days until/since ddate	
				remdays = None
				if bset and today >= b:
					if begindueset:
						remdays = (due - today).days
					else:
						remdays = (b - today).days
				else:
					if begindueset and today >= begin:
						remdays = (due - today).days
										
				if complete:
					remdays = "X"
				elif remdays:
					remdays += 1
					remdays = "%s" % remdays
				else:
					remdays = ''
	
				if len(a) > actwidth:
					actt = "%-*s..." % (actwidth-3, a[:actwidth-3])
				else:
					actt = a

				ai += 1
				i = "%s/%s" % (ai, num_actions)
					
				if c == "???" or agenda == "Inbox":
					if delimiter != '':
						if f:
							z = f.strftime(delfmt)
						else:
							z = ''
						l = [file, agenda, i, c, b.strftime(delfmt), 
							"%s" % d.days, z, 
							ddate.strftime(datefmt), remdays, a]
						s = delimiter.join(l)
					else:
						s = "%-*s %-*s %-6s %-3s %-5s %-*s" % (
							agendawidth, agendat, cwidth, 
							c, ddate.strftime(datefmt),
							 remdays, i, actwidth, actt)
						l = [ct, ddate, i, s] 
					inbox.append(l)
					continue
					
				if remtemplate and remind:
					remhash = {}
					addrem = True
					if not complete and b <= today:
						remplist.append(p)
						if len(remplist) > 1:
								if remplist[-1] != remplist[-2]:
									addrem = False
								elif series and (remplist[-1] 
									== '' and remplist[-2] == ''):
									addrem = False
						if addrem:
							# remtemplate = 'REM %(d)s %(r)s MSG
							# %%"%(p)s/%(c)s %(a)s%%" %%1%%'
							remhash['d'] = today.strftime("%b %d %Y")							
							remhash['r'] = remind
							remhash['p'] = agendat
							remhash['c'] = c
							remhash['a'] = actt
							reminder = remtemplate % remhash
							fo = open(remfile, 'a')
							fo.write("%s\n" % reminder)
							fo.close()

				if filter == "remaining":
					if complete: 
						continue
				elif filter == "completed":
					if not complete:
						continue
				elif filter == "available":
					if complete:
						continue
					elif b > today:
						continue
					else:
						plist.append(p)
					if len(plist) > 1:
							if plist[-1] != plist[-2]:
								break
							elif series and \
								(plist[-1] == '' and plist[-2] == ''):
								break
				if delimiter != '':
					if f:
						z = f.strftime(delfmt)
					else:
						z = ''
					l = [file, agenda, i, c, b.strftime(delfmt), 
						"%s" % d.days, z, ddate.strftime(datefmt),
						remdays, a]
					s = delimiter.join(l)
				else:
					s = " %-*s %-*s %-6s %-3s %-5s %-*s" % (
						agendawidth, agendat, cwidth, 
						c, ddate.strftime(datefmt),
						 remdays, i, actwidth, actt)
				if view == "context":
					l = [ct, ddate, i, s, file, ai] 
					acts.append(l)
				elif view == "project":
					l = [agendat, i, ddate, s, file, ai] 
					acts.append(l)
				elif view == "date":
					l = [ddate, agendat, i, s, file, ai] 
					acts.append(l)
			if not unfinished and repeating:
				proj['lastdone'] = lastcompleted
				proj['lastdue'] = due
				saveFileData(file, proj)
				logaction(file, "set lastdue = %s and lastdone = %s" %
					 (due, lastcompleted))
					
	if errors:
		for error in errors:
			print error
		sys.exit()

	if len(inbox) > 0:
		inbox.sort()
		if delimiter == '':
			str = " Inbox - needs clearing "
			lbar = (71-len(str))/2
			bars = "="*lbar
			output += "%s%s%s\n" %  (bars, str, bars)
		for act in inbox:
			output += "%s\n" % act[3].strip()

	if len(acts) > 0:
		acts.sort()
		# for act in acts:
		# 	print act
		index = 0
		if delimiter == '':
			str = " %s actions by %s " % (filter, view)
			lbar = (71-len(str))/2
			bars = "="*lbar
			output += "%s%s%s\n" %  (bars, str, bars)
		for act in acts:
			if act[3].strip():
				if mark:
					index += 1
					output += "%2s %s\n" % (index,act[3].strip())
					fileinfo.append([act[4],act[5]])
				else:
					output += "%s\n" % act[3].strip()

	if printoutput:
		fo = open(print_file, 'w')
		fo.write(output)
		fo.close()
		os.system("%s %s" % (print_cmd, print_file))
	elif ret:
		return output
	elif output:
		print "%s" % output
		return fileinfo
		
def main():
	parser = OptionParser(version = "yeagtd %s" % version,
		description = "Produce nicely formatted summary reports of simple GTD project files. Repetition is supported and tasks can be sorted by context, project or date and filtered in a variety of ways. Output can optionally be printed or exported in either comma or tab delimited format.")
	parser.add_option("-s", "--sort", action = "store", type="string",
		dest='sort', default = "c", 
		help = """Sort by [c]ontext, [p]roject or [d]ue date. Defaults to [c]ontext.""")
	parser.add_option("-f", "--filer", action = "store", type="string", dest='filter', 
		default="a", help = """Impose [n]o filters or limit actions shown to those which are [c]ompleted, [r]emaining or [a]vailable. Defaults to [a]vailable.""")
	parser.add_option("-x", "--export", action = "store", type="string", 
		dest='export', 
		default="", help = """Export actions using sort and filter settings to [p]rinter or to standard output in [c]omma delimited or [t]ab delimited format. Defaults to [t]ab.""")
	parser.add_option("-b", "--backup", 
		action="callback", callback = backup,
		help="Backup all project files.")
	parser.add_option("-d", "--done", 
		action="callback", callback = recorddone,
		help="Choose action(s) to mark completed from from those which are available.")
	parser.add_option("-e", "--edit", action = "store_true",
		dest='edit',
		help = """Choose file(s) to open for editing from existing project files.""")
	parser.add_option("-a", "--add", action = "store_true",
		dest='add',
		help = """Ignore other switches and add an un-categorized action to the inbox using the provided argument as the title.""")
	parser.add_option("-t", "--topic", action = "store_true",
		dest='topic',
		help = """Ignore other switches and create a new project file using the provided argument as the relative path.""")
	parser.add_option("-H", "--Help", action = "store_true",
		dest='cheatsheet',
		help = """Ignore other switches and show a project file 'CheatSheet'.""")
	(options, args) = parser.parse_args()
	if options.cheatsheet:
		CheatSheet()
	if options.add:
		inbox()
	elif options.edit:
		edit()
	elif options.topic:
		topic()
	else:
		data = loadData()
		parseData(data, options.sort, options.filter, options.export)

if __name__ == '__main__':
	main()